/* 
 * Process Variable Library with JVisu Socket Server
 * 
 * Copyright (C) 2003 Thomas Gleixner <tglx@linutronix.de>
 *           (C) 2005 Robert Schwebel <r.schwebel@pengutronix.de>
 *		
 * PURPOSE
 * 
 * This is a lightweight, single threaded multiconnection server for
 * data visualization based on JVisu applets. 
 * 
 * It's single threaded as it is intended to run on as much platforms
 * as possible. For uClinux a single threaded server is the best choice.
 * On standard linux platforms this concept does not hurt.
 * 
 * All you have to provide is a structure with functions for symbol 
 * resolving and variable write. Variable write can be NULL, get_symbol 
 * must be a valid function.
 * 
 * The server is running as long as the global variable jvisu_stop is zero.
 * You can stop the server by modifying jvisu_stop.
 * Signal handling is not done inside this code to make it as portable as
 * possible. You can do this in the thread which calls this code.
 *
 * Note that most communication related data is wrapped with special 
 * datatypes which let you not access this data directly. They ensure
 * endianess conversion. Use the appropriate conversion macro.
 */

#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <limits.h>
#include <sys/sem.h>
#include <sys/ipc.h>
#include <time.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <expat.h>
#include <libgen.h>
#include <pthread.h>
#include <ptxlist.h>

#include <pv.h>

#define D1(x) if (pv_verbose) x
static int pv_verbose = 0;

/* private prototypes */
static void unregister_vars (pv_engine_t *pv_engine, struct jvisu_sock *sd);
static void shutdown_sock (pv_engine_t *pv_engine, struct jvisu_sock *sd);
static void check_variables (pv_engine_t *pv_engine);
static int  start_socket (pv_engine_t *pv_engine, int port);
static int  do_accept (pv_engine_t *pv_engine, int sd);
static int  analyse_command (pv_engine_t *pv_engine, struct jvisu_sock *sd);
static int  process_request (pv_engine_t *pv_engine, struct jvisu_sock *sd);
static void check_fdset (pv_engine_t *pv_engine);
static int  get_symbol (pv_engine_t *pv_engine, struct jvisu_var *var, char *symbol);

static const char *prompt = "jvisu";

/* xml stuff */

typedef enum {
	tag_undef = 0,
	tag_name,
	tag_type,
	tag_offset,
	tag_desc,
	tag_default,
	tag_unit,
} xml_tag_t;

#ifdef _SEM_SEMUN_UNDEFINED
union semun {
	int val;
	struct semid_ds *buf;
	unsigned short int *array;
	struct seminfo *__buf;
};
#endif

/** 
 *	Unregister all variables which are registered for a given 
 *	socket connection
 *  
 *	@param sd	socket description structure
 */
static void unregister_vars (pv_engine_t *pv_engine, struct jvisu_sock *sd)
{
	struct jvisu_var *var;
	int finished = 0; 

	if (is_empty(pv_engine->jv_varlist))
		return;
	
	/* remove from list, free memory */
	for (goto_head(&pv_engine->jv_varlist); !finished; goto_next(&pv_engine->jv_varlist)) {
		
		var = (struct jvisu_var*)list_get_element_data(pv_engine->jv_varlist);

		if (is_last(pv_engine->jv_varlist))
			finished=1;
		
		if (var->sd->sd != sd->sd)
			continue;
		
		pv_engine->jv_registeredvars--;
		pv_engine->jv_varlist = list_del_element(pv_engine->jv_varlist);
	}
}


/** 
 * 	Shut down a socket and unregister variables
 * 	
 *	@param sd	socket description structure
 */
static void shutdown_sock (pv_engine_t *pv_engine, struct jvisu_sock *sd)
{
	D1 (printf ("%s: %-20s: %d\n", prompt, __FUNCTION__, sd->sd));

	unregister_vars(pv_engine, sd);
	pv_engine->jv_connections--;
	FD_CLR (sd->sd, &pv_engine->jv_read_fdset);
	FD_CLR (sd->sd, &pv_engine->jv_write_fdset);
	close (sd->sd);
	sd->sd = -1;
}


/** 
 *	Retrieve a boolean variable from the application; do typecasting 
 *	from internal format
 *
 *	@param var	variable description structure
 *	@return		typecasted variable content 
 */
static inline u_char get_bool (struct jvisu_var *var)
{
	switch (var->inttype) {
		case jv_bool:
			return (*((u_char *) var->data) & var->bitmask) ? 1 : 0;
		case jv_short:
			return (*((uint16_t *) var->data)) ? 1 : 0;
		case jv_int:
			return (*((uint32_t *) var->data)) ? 1 : 0;
		case jv_float:
			return (*((float *) var->data)) ? 1 : 0;
		case jv_double:
			return (*((double *) var->data)) ? 1 : 0;
		case jv_string:	
		case jv_byte:
			return (*((u_char *) var->data)) ? 1 : 0;
		default:
			break;
	}	
	return 0;
}


/** 
 *	Retrieve a byte variable from the application; do typecasting 
 *	from internal format
 *	
 *	@param var	variable description structure
 *	@return		typecasted variable content 
 */
static inline u_char get_byte (struct jvisu_var *var)
{
	switch (var->inttype) {
		case jv_bool:
			return (*((u_char *) var->data) & var->bitmask) ? 1 : 0;
		case jv_short:
			return (u_char) (*((uint16_t *) var->data));
		case jv_int:
			return (u_char) (*((uint32_t *) var->data));
		case jv_float:
			return (u_char) (*((float *) var->data));
		case jv_double:
			return (u_char) (*((double *) var->data));
		case jv_string:	
		case jv_byte:
			return (*((u_char *) var->data));
		default:
			break;
	}	
	return 0;
}


/** 
 *	Retrieve a short variable from the application; do typecasting
 *	from internal format
 *	
 *	@param var	variable description structure
 *	@return		typecasted variable content 
 */
static inline short get_short (struct jvisu_var *var)
{
	switch (var->inttype) {
		case jv_bool:
			return (*((u_char *) var->data) & var->bitmask) ? 1 : 0;
		case jv_short:
			return (*((uint16_t *) var->data));
		case jv_int:
			return (short) (*((uint32_t *) var->data));
		case jv_float:
			return (short) (*((float *) var->data));
		case jv_double:
			return (short) (*((double *) var->data));
		case jv_string:	
			return (short) atoi ((u_char *) var->data);
		case jv_byte:
			return (short) (*((u_char *) var->data));
		default:
			break;
	}	
	return 0;
}


/** 
 *	Retrieve a integer variable from the application; do typecasting 
 *	from internal format
 *	
 *	@param var	variable description structure
 *	@return		typecasted variable content 
 */
static inline int get_int (struct jvisu_var *var)
{
	switch (var->inttype) {
		case jv_bool:
			return (*((u_char *) var->data) & var->bitmask) ? 1 : 0;
		case jv_short:
			return (int) (*((uint16_t *) var->data));
		case jv_int:
			return (*((uint32_t *) var->data));
		case jv_float:
			return (int) (*((float *) var->data));
		case jv_double:
			return (int) (*((double *) var->data));
		case jv_string:	
			return atoi ((u_char *) var->data);
		case jv_byte:
			return (int) (*((u_char *) var->data));
		default:
			break;
	}	
	return 0;
}


/** 
 *	Retrieve a float variable from the application; do typecasting 
 *	from internal format
 *
 *	@param var	variable description structure
 *	@return		typecasted variable content 
 */
static inline float get_float (struct jvisu_var *var)
{
	switch (var->inttype) {
		case jv_bool:
			return (*((u_char *) var->data) & var->bitmask) ? 1 : 0;
		case jv_short:
			return (float) (*((uint16_t *) var->data));
		case jv_int:
			return (float) (*((uint32_t *) var->data));
		case jv_float:
			return (*((float *) var->data));
		case jv_double:
			return (float) (*((double *) var->data));
		case jv_string:	
			return (float) atof ((u_char *) var->data);
		case jv_byte:
			return (float) (*((u_char *) var->data));
		default:
			break;
	}	
	return 0;
}


/** 
 *	Retrieve a double variable from the application; do typecasting 
 *	from internal format
 *	
 *	@param var	variable description structure
 *	@return		typecasted variable content 
 */
static inline double get_double (struct jvisu_var *var)
{
	switch (var->inttype) {
		case jv_bool:
			return (*((u_char *) var->data) & var->bitmask) ? 1 : 0;
		case jv_short:
			return (double) (*((uint16_t *) var->data));
		case jv_int:
			return (double) (*((uint32_t *) var->data));
		case jv_float:
			return (double) (*((float *) var->data));
		case jv_double:
			return (*((double *) var->data));
		case jv_string:	
			return atof ((u_char *) var->data);
		case jv_byte:
			return (double) (*((u_char *) var->data));
		default:
			break;
	}	
	return 0;
}


/** 
 *	Check if variables have been added or changed
 */
static void check_variables (pv_engine_t *pv_engine)
{
	struct timeval		tnow, tdiff;
	int			ms;
	int			finished = 0;

	if (is_empty(pv_engine->jv_varlist))
		return; 

	/* read current time */
	gettimeofday (&tnow, NULL);
	tdiff.tv_sec = tnow.tv_sec - pv_engine->jv_tact.tv_sec;
    	tdiff.tv_usec = tnow.tv_usec - pv_engine->jv_tact.tv_usec;
	if (tdiff.tv_usec < 0) {
		tdiff.tv_sec--;
		tdiff.tv_usec += 1000000;
	}

	/* convert to ms */
	ms = (tdiff.tv_sec * 1000) + (tdiff.tv_usec / 1000);
	if (ms <= 0) {
		printf("0");
		return;
	}

	/* save current time and take care of the usec remainder */
	pv_engine->jv_tact = tnow;
	pv_engine->jv_tact.tv_usec += tdiff.tv_usec % 1000;
	if (pv_engine->jv_tact.tv_usec >= 1000000) {
		pv_engine->jv_tact.tv_sec++;
		pv_engine->jv_tact.tv_usec -= 1000000;
	}

	/* walk through the list */
	for (goto_head(&pv_engine->jv_varlist); !finished; goto_next(&pv_engine->jv_varlist)) {
		
		size_t			len;
		int 			res, force;
		struct jvisu_sndvar	*snd;
		struct jvisu_var 	*var = (struct jvisu_var*)list_get_element_data(pv_engine->jv_varlist);
		var->elapsed += ms;

		if (is_last(pv_engine->jv_varlist))
			finished=1;
		
		/* time elapsed ? */
		if (var->elapsed < var->rate)
			continue;

		/* transfer in progress ? */
		if (var->sd->status || (var->permissions & JVISU_PERM_WO)) 
			continue;

		snd = (struct jvisu_sndvar *) var->sd->buf;
		snd->hdr.cmd = cmd_send_var;
		snd->hdr.type = (u_char) var->type;
		len = sizeof (struct jvisu_sndvar);
		snd->idx = cpu_to_je32 (var->idx);
		
		force = (var->status & JVISU_STAT_INIT) ? 0 : 1;
		
		switch (var->type) {
		case jv_byte: {
			pv_lock(pv_engine);
			u_char act = get_byte (var);
			pv_unlock(pv_engine);
			if ((var->cmp.b) == act && !force)
				continue;
			var->cmp.b = act;
			len++;
			snd->data.b[0] = act;
			break;
		}	
		case jv_short: {
			pv_lock(pv_engine);
			short act = get_short (var);
			pv_unlock(pv_engine);
			if ((var->cmp.s) == act && !force)
				continue;
			var->cmp.s = act;
			len += sizeof (short);
			snd->data.s[0] = cpu_to_je16 (act);
			break;
		}	
		case jv_int: {
			pv_lock(pv_engine);
			int act = get_int (var);
			pv_unlock(pv_engine);
			if ((var->cmp.i) == act && !force)
				continue;
			var->cmp.i = act;
			len += sizeof (int);
			snd->data.i[0] = cpu_to_je32 (act);
			break;
		}	
		case jv_float: {
			union { float f; uint32_t i; } act;
			pv_lock(pv_engine);
			act.f = get_float (var);
			pv_unlock(pv_engine);
			if ((var->cmp.f) == act.f && !force)
				continue;
			var->cmp.f = act.f;
			len += sizeof (float);
			snd->data.i[0] = cpu_to_je32 (act.i);
			break;
		}	
		case jv_double: {
			union { double d; uint64_t i; } act;
			pv_lock(pv_engine);
			act.d = get_double (var);
			pv_unlock(pv_engine);
			if ((var->cmp.d) == act.d && !force)
				continue;
			var->cmp.d = act.d;
			len += sizeof (double);
			snd->data.d[0] = cpu_to_je64 (act.i);
			break;
		}	
		case jv_bool: {
			pv_lock(pv_engine);
			u_char act = get_bool (var);
			pv_unlock(pv_engine);
			if ((var->cmp.b == act) && !force)
				continue;
			var->cmp.b = act;
			len++;
			snd->data.b[0] = act;
			break;
		}
			      
		/* 
		 * TYPE String provides no typecasting ATM
		 * As I have no real good idea to track changes on strings
		 * due to variable length and content I decided to transfer strings 
		 * in the given interval. 
		 * If you want to have change triggered string display 
		 * then use a index variable and provide the strings in a file
		 * which can be retrieved by the applet 
		 * If you are still not happy provide a better solution :)
		 */

		case jv_string:
			len += strlen ((char *)var->data);
			strcpy (snd->data.b, (char *)var->data);
			break;

		/* 
		 * TYPE Stream is a special case. Check, if the length variable
		 * is > 0. Send out the bytes and set it to zero
		 */
			
		case jv_stream: {
			struct jvisu_varstream *p = (struct jvisu_varstream*) var->data;
			if (!p->len)
				continue;
			len += p->len;
			D1 (
				printf ("%s: %-20s: jv_stream: Send %d bytes (total %d)\n", 
				prompt, __FUNCTION__, p->len, len)
			);
			memcpy (snd->data.b, p->data, (size_t)p->len);
			p->len = 0;
			break;
		}

		case jv_bytearray: {
			int i;
			char *p = (char *) var->data;
			for (i = 0; i < var->arraylen; i++) {
				if (p[i] != var->cmp.p[i]) {
					force = 1;
					break;
				}
			}		
			if (!force)
				continue;
			memcpy (snd->data.b, var->data, (size_t) var->arraylen);
			len += var->arraylen;
			break;
		}

		default:
			continue;
		}

		/* send value */
		snd->hdr.len = cpu_to_je16 ((ushort) len);
		res = write (var->sd->sd, var->sd->buf, len);
		if (res < 0) {
			if (res != EAGAIN) {
				shutdown_sock(pv_engine, var->sd);
				continue;
			}
			res = 0;
		}
		/* all sent ? */
		if (res == len) {
			var->elapsed = 0;	
			var->status |= JVISU_STAT_INIT;
			continue;
		}
		fprintf (stderr, "%s: %-20s: FIXME: Send has to be delayed\n", prompt, __FUNCTION__);	
		var->sd->bufpoi = res;
		var->sd->cmdlen = len;
		var->sd->status	= STAT_WRITE;

	}	
}


/** 
 *	Write a boolean variable in the application; do typecasting 
 *	from applet format
 *	
 *	@param var	variable description structure
 *	@param dat	union containing the data to write
 */
static inline void set_bool (struct jvisu_var *var, union jvisu_pv *pv)
{
	int res;

	switch (var->type) {
		case jv_bool:
		case jv_string:	
		case jv_byte:
			res = pv->b[0] ? 1 : 0;
			break;
		case jv_short:
			res = pv->s[0] ? 1 : 0;
			break;
		case jv_int:
			res = pv->i[0] ? 1 : 0;
			break;
		case jv_float:
			res = (pv->f[0] != 0.0) ? 1 : 0;
			break;
		case jv_double:
			res = (pv->d[0] != 0.0) ? 1 : 0;
			break;
		default:
			return;
	}
	if (res)
		*((u_char *) var->data) |= var->bitmask;
	else	
		*((u_char *) var->data) &= ~var->bitmask;
}


/** 
 *	Write a byte variable in the application; do typecasting 
 *	from applet format
 *	
 *	@param var	variable description structure
 *	@param dat	union containing the data to write
 */
static inline void set_byte (struct jvisu_var *var, union jvisu_pv *pv)
{

	switch (var->type) {
		case jv_bool:
		case jv_byte:
			*((u_char *) var->data) = pv->b[0];
			break;
		case jv_short:
			*((u_char *) var->data) = (u_char) pv->s[0];
			break;
		case jv_int:
			*((u_char *) var->data) = (u_char) pv->i[0];
			break;
		case jv_float:
			*((u_char *) var->data) = (u_char) pv->f[0];
			break;
		case jv_double:
			*((u_char *) var->data) = (u_char) pv->d[0];
			break;
		case jv_string:
			*((u_char *) var->data) = (u_char) atoi (pv->b);
			break;
		default:
			return;
	}
}


/** 
 *	Write a short variable in the application; do typecasting 
 *	from applet format
 *	
 *	@param var	variable description structure
 *	@param dat	union containing the data to write
 */
static inline void set_short (struct jvisu_var *var, union jvisu_pv *pv)
{

	switch (var->type) {
		case jv_bool:
		case jv_byte:
			*((short *) var->data) = (short) pv->b[0];
			break;
		case jv_short:
			*((short *) var->data) = pv->s[0];
			break;
		case jv_int:
			*((short *) var->data) = (short) pv->i[0];
			break;
		case jv_float:
			*((short *) var->data) = (short) pv->f[0];
			break;
		case jv_double:
			*((short *) var->data) = (short) pv->d[0];
			break;
		case jv_string:
			*((short *) var->data) = (short) atoi (pv->b);
			break;
		default:
			return;
	}
}


/** 
 *	Write a integer variable in the application; do typecasting 
 *	from applet format
 *	
 *	@param var	variable description structure
 *	@param pv	union containing the data to write
 */
static inline void set_int (struct jvisu_var *var, union jvisu_pv *pv)
{

	switch (var->type) {
		case jv_bool:
		case jv_byte:
			D1 (
				printf ("%s: %-20s: var data: %d\n", 
				prompt, __FUNCTION__, (int) pv->b[0])
			);
			*((int *) var->data) = (int) pv->b[0];
			break;
		case jv_short:
			*((int *) var->data) = (int) pv->s[0];
			break;
		case jv_int:
			*((int *) var->data) = pv->i[0];
			break;
		case jv_float:
			*((int *) var->data) = (int) pv->f[0];
			break;
		case jv_double:
			*((int *) var->data) = (int) pv->d[0];
			break;
		case jv_string:
			*((int *) var->data) = atoi (pv->b);
			break;
		default:
			return;
	}
}


/** 
 *	Write a float variable in the application; do typecasting 
 *	from applet format
 *	
 *	@param var	variable description structure
 *	@param pv	union containing the data to write
 */
static inline void set_float (struct jvisu_var *var, union jvisu_pv *pv)
{

	switch (var->type) {
		case jv_bool:
		case jv_byte:
			*((float *) var->data) = (float) pv->b[0];
			break;
		case jv_short:
			*((float *) var->data) = (float) pv->s[0];
			break;
		case jv_int:
			*((float *) var->data) = (float) pv->i[0];
			break;
		case jv_float:
			*((float *) var->data) = pv->f[0];
			break;
		case jv_double:
			*((float *) var->data) = (float) pv->d[0];
			break;
		case jv_string:
			*((float *) var->data) = (float) atof (pv->b);
			break;
		default:
			return;
	}
}


/** 
 *	Write a double variable in the application; do typecasting 
 *	from applet format
 *	
 *	@param var	variable description structure
 *	@param pv	union containing the data to write
 */
static inline void set_double (struct jvisu_var *var, union jvisu_pv *pv)
{

	switch (var->type) {
		case jv_bool:
		case jv_byte:
			*((double *) var->data) = (double) pv->b[0];
			break;
		case jv_short:
			*((double *) var->data) = (double) pv->s[0];
			break;
		case jv_int:
			*((double *) var->data) = (double) pv->i[0];
			break;
		case jv_float:
			*((double *) var->data) = (double) pv->f[0];
			break;
		case jv_double:
			*((double *) var->data) = pv->d[0];
			break;
		case jv_string:
			*((double *) var->data) = atof (pv->b);
			break;
		default:
			return;
	}
}


/** 
 *	Write a string variable in the application; do typecasting 
 *	from applet format
 *	
 *	@param var	variable description structure
 *	@param pv	union containing the data to write
 */
static inline void set_string (struct jvisu_var *var, union jvisu_pv *pv)
{

	switch (var->type) {
	case jv_bool:
	case jv_byte:
		sprintf ((char *) var->data, "%d", (int) pv->b[0]);
		break;
	case jv_short:
		sprintf ((char *) var->data, "%d", (int) pv->s[0]);
		break;
	case jv_int:
		sprintf ((char *) var->data, "%d", pv->i[0]);
		break;
	case jv_float:
		sprintf ((char *) var->data, "%f", pv->f[0]);
		break;
	case jv_double:
		sprintf ((char *) var->data, "%f", pv->f[0]);
		break;
	case jv_string:
		strcpy ((char *) var->data, pv->b);
		break;
	default:
		return;
	}
}


/** 
 *	Find a matching variable and endianess-convert the incoming data, 
 *	then call the appropriate write function depending on internal
 *	variable type; call the variable specific write function if available
 *
 *	@param pv	receive buffer structure
 *	@param sd	socket descriptor on which the write command was received
 */
static void write_variable (pv_engine_t *pv_engine, struct jvisu_sndvar	*dat, struct jvisu_sock *sd)
{
	char			buf[MAX_DATA_LEN];
	union jvisu_pv	 	*idat = (union jvisu_pv *) buf;
	int 			finished = 0; 
	
	D1(
		printf ("%s: %-20s: idx = %d\n", prompt, __FUNCTION__, je32_to_cpu(dat->idx))
	);	

	if (is_empty(pv_engine->jv_varlist)) {
		D1(printf("%s: %-20s: no registered variables\n", prompt, __FUNCTION__));
		return;
	}
	
	/* find the matching entry */	
	for (goto_head(&pv_engine->jv_varlist); !finished; goto_next(&pv_engine->jv_varlist)) {

		struct jvisu_var *var = (struct jvisu_var*)list_get_element_data(pv_engine->jv_varlist);

		if (is_last(pv_engine->jv_varlist))
			finished=1;
		
		if (var->idx != je32_to_cpu (dat->idx) || var->sd != sd)
			continue;
			
		/* check, if this variable is allowed to be written */
		if ((var->permissions & JVISU_PERM_RO)) {
			D1 (printf ("%s: %-20s: request to write readonly var!\n", prompt, __FUNCTION__));
			return;
		}	
		D1 (printf ("%s: %-20s: type: %d, %d\n", prompt, __FUNCTION__, var->type, var->inttype));	
		
		/* do endianess conversion */	
		switch (var->type) {
			case jv_byte:
			case jv_bool:
				idat->b[0] = dat->data.b[0];
				break;
			case jv_short:
				idat->u16[0] = je16_to_cpu (dat->data.s[0]);
				break;
			case jv_int:
			case jv_float:
				idat->u32[0] = je32_to_cpu (dat->data.i[0]);
				break;
			case jv_double: {
				idat->u64[0] = je64_to_cpu (dat->data.d[0]);
				break;
			}	
			case jv_string:
				dat->data.b[(sd->cmdlen - sizeof (struct jvisu_cmdhdr))] = '\0';
				strcpy (idat->b, dat->data.b);
				break;
			case jv_stream:
				break;	
			case jv_bytearray: 
				break;	
			default:
				return;	
		}

		/* call the proper write function*/	
		switch (var->inttype) {
			case jv_bool:
				pv_lock(pv_engine);
				set_bool (var, idat); 
				pv_unlock(pv_engine);
				break;;
			case jv_byte:
				pv_lock(pv_engine);
				set_byte (var, idat); 
				pv_unlock(pv_engine);
				break;
			case jv_short:
				pv_lock(pv_engine);
				set_short (var, idat); 
				pv_unlock(pv_engine);
				break;
			case jv_int:
				pv_lock(pv_engine);
				set_int (var, idat); 
				pv_unlock(pv_engine);
				break;
			case jv_float:
				pv_lock(pv_engine);
				set_float (var, idat); 
				pv_unlock(pv_engine);
				break;
			case jv_double:
				pv_lock(pv_engine);
				set_double (var, idat); 
				pv_unlock(pv_engine);
				break;
			case jv_string:
				pv_lock(pv_engine);
				set_string (var, idat); 
				pv_unlock(pv_engine);
				break;
			case jv_stream: {
				struct jvisu_varstream *p = (struct jvisu_varstream*) var->data;
				pv_lock(pv_engine);
				memcpy (p->data, dat->data.b, (sd->cmdlen - sizeof (struct jvisu_cmdhdr)));
				pv_unlock(pv_engine);
				p->len = (sd->cmdlen - sizeof (struct jvisu_cmdhdr));
				break;
				}	
			case jv_bytearray: {
				/* check length */
				int num = sd->cmdlen - sizeof (struct jvisu_cmdhdr);
				if (num > var->arraylen)
					num = var->arraylen;
				pv_lock(pv_engine);
				memcpy (var->data, dat->data.b, (size_t) num);
				pv_unlock(pv_engine);
				D1 (printf ("%s: %-20s: jv_bytearray: len = %d\n", prompt, __FUNCTION__, num));
				break;
				}
		}
		
		/* check if a write_done function for this variable is available */
// RSC
//		if (var->write_done)
//			var->write_done (var->arg);
		
		var->elapsed += var->rate;	
	}
}


/** 
 *	Start socket
 *	
 *	@param port	portnumber to listen
 *	@return		filedescriptor or errorcode
 */
static int start_socket (pv_engine_t *pv_engine, int port)
{
	int			sd, opt;
	struct sockaddr_in 	sin;	

	/* open the socket */
	sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (sd < 0)
		return sd;
		
	/* set non blocking */
	if (fcntl(sd, F_SETFL, O_NONBLOCK) < 0) 
		goto err;

	/* close on exit */
	if (fcntl(sd, F_SETFD, 1) < 0)
		goto err;
	
	/* reuse address */
	opt = 1;
	if (setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, (void *) &opt, sizeof(opt)) < 0)
		goto err;

	/* fill in the neccecary stuff */
	memset(&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = htonl (INADDR_ANY);
	sin.sin_port = htons ((uint16_t) port);		

	/* bind it */
	if (bind (sd, (struct sockaddr *) &sin, sizeof (sin)) < 0)
		goto err;
	
	/* try to listen with max connections */
	if (listen(sd, MAX_CONNECTIONS) < 0)
		goto err;

	return sd;
	
err:
	fprintf (stderr, "%s: %-20s: error\n", prompt, __FUNCTION__);
	close (sd);
	return -1;
}


/** 
 *	Accept connections if we have less than MAX_CONNECTIONS
 *	
 *	@param sd	socket descriptor
 *	@return		new socket descriptor or error
 */
static int do_accept (pv_engine_t *pv_engine, int sd)
{
	int 			i, fd, opt = 1;
	struct sockaddr_in 	raddr;
	int			addrlen = sizeof (raddr);

	D1 (printf ("%s: %-20s: accepting connection\n", prompt, __FUNCTION__));

	/* sorry, enough connections */
	if (pv_engine->jv_connections > MAX_CONNECTIONS) {
		D1(
			printf ("%s: %-20s: error: more than MAX_CONNECTIONS open\n", prompt, __FUNCTION__)
		);
		return 0;
	}

	/* try to accept the connection */
  	raddr.sin_family = 0xdead;
	fd = accept(sd, (struct sockaddr *) &raddr, &addrlen);

	if (fd < 0) {
		D1 (printf ("%s: %-20s: error: failed!\n", prompt, __FUNCTION__));
		if (errno == EAGAIN || errno == EWOULDBLOCK)
			return 0;
		else 	
			return fd;
	}

	/* some linux kernels have this problem when client closes between
	 * select() and accept()	
	 */
	if (raddr.sin_family != AF_INET)
		goto err;

	/* keep alive */
	if (setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, (void *) &opt, sizeof(opt)) < 0)
		goto err;
	/* reuse address */
	if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (void *) &opt, sizeof(opt)) < 0)
		goto err;
	/* no delay */
	if (setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, (void *) &opt, sizeof (opt)) < 0)
		goto err;
	
	/* non blocking mode */
	if (fcntl (fd, F_SETFL, O_NONBLOCK) < 0) 
		goto err;

	/* close on exit */
	if (fcntl(fd, F_SETFD, 1) < 0)
		goto err;

	/* put it into an empty place (-1 => empty) */
	for (i = 0; i <= MAX_CONNECTIONS; i++) {
		if (pv_engine->jv_active_sock[i].sd == -1) {
			pv_engine->jv_active_sock[i].sd = fd;
			pv_engine->jv_active_sock[i].status = 0;
			pv_engine->jv_active_sock[i].bufpoi = 0;
			pv_engine->jv_active_sock[i].cmdlen = 0;
			pv_engine->jv_connections++;
			D1(
				printf ("%s: %-20s: socket descriptor stored: sd=%d, currently %d connection%s active\n", 
				prompt, __FUNCTION__, fd, pv_engine->jv_connections, pv_engine->jv_connections==1 ? "" : "s")
			);
			return 0;	
		}
	}
	/* should not happen */
	pv_engine->jv_connections = MAX_CONNECTIONS;
	fprintf (stderr, "%s: %-20s: eeep! no free activ_sock, but connections < max\n", prompt, __FUNCTION__);
		
	/* error exit for valid socket descriptors */
err:
	fprintf (stderr, "%s: %-20s: error occured, exiting\n", prompt, __FUNCTION__);
	close(fd);
	return 0;
}


/** 
 *	Analyse command received via JVisu Native Protocol
 *	
 *	@param sd	jvisu socket description structure
 *	@return		0 on success, else error number
 */
static int analyse_command (pv_engine_t *pv_engine, struct jvisu_sock *sd)
{
	int	res = 0;

	switch (sd->buf[0]) {
	
	/* 
	 * the applets wants to register a variable with us
	 * FIXME: handle string variables
	 */

	case cmd_register_var: {

		struct jvisu_var var;
		struct jvisu_regvar_cmd	*reg_cmd = (struct jvisu_regvar_cmd *) sd->buf;
		int namelen;

		/* truncate symbol, if neccecary */
		namelen = MIN (sd->bufpoi - sizeof (struct jvisu_regvar_cmd), MAX_SYMBOL_LEN);

		/* terminate string */
		sd->buf[sd->cmdlen] = 0x0;

		/* copy data from network command into jvisu_var */
		var.sd = sd;
		var.type = (int)reg_cmd->hdr.type;
		var.idx = je32_to_cpu(reg_cmd->idx);
		var.rate = je32_to_cpu(reg_cmd->rate);
		var.elapsed = var.rate;
		
		D1(
			printf ("%s: %-20s: cmd_register_var(%s), idx=%d\n", 
			prompt, __FUNCTION__, reg_cmd->symbol, var.idx)
		);

		/* get a pointer to the variable; if it's a bool, get a bitmask too */
		D1(printf("%s: %-20s: call jobs->get_symbol()\n", prompt, __FUNCTION__));

		if (get_symbol (pv_engine, &var, reg_cmd->symbol) < 0) {
			res = -EINVAL;
			break;
		}
		/* for byte arrays get the arraysize, too */
		if (var.inttype == jv_bytearray)
			var.cmp.p = malloc ((size_t)var.arraylen);
			
		/* add var to end of jv_varlist */
		if (pv_engine->jv_varlist == NULL)
			pv_engine->jv_varlist = list_init();

		pv_engine->jv_varlist = 
			list_add_tail(pv_engine->jv_varlist, (void*)&var, sizeof(struct jvisu_var));

		pv_engine->jv_registeredvars++;
		break;
	}
	
	/* FIXME: Handle string variable */
	case cmd_unregister_var: {

		struct jvisu_regvar_cmd *reg = (struct jvisu_regvar_cmd *) sd->buf;
		int finished = 0; 
		
		D1(
			printf("%s: %-20s: cmd_unregister_var(%s), idx = %d\n", 
			prompt, __FUNCTION__, reg->symbol, je32_to_cpu (reg->idx))
		);
			
		if (is_empty(pv_engine->jv_varlist))
			return -EINVAL; /* FIXME: correct? */
		
		/* remove from list, free memory */
		for (goto_head(&pv_engine->jv_varlist); !finished; goto_next(&pv_engine->jv_varlist)) {

			struct jvisu_var *var = 
				(struct jvisu_var*)list_get_element_data(pv_engine->jv_varlist);
		
			if (is_last(pv_engine->jv_varlist))
				finished=1;

			if (var->idx != je32_to_cpu (reg->idx) || var->sd != sd) {
				continue;
			}
			
			pv_engine->jv_registeredvars--;
			
			if (var->inttype == jv_bytearray)
				if (var->cmp.p)
					free(var->cmp.p);

			pv_engine->jv_varlist = list_del_element(pv_engine->jv_varlist);
			break;	/* break here, as it is unique (hopefully...) */
		}
		
		break;
	}
				 
	case cmd_conf_data: {
				    
		struct jvisu_sndgen *dat = (struct jvisu_sndgen *) sd->buf;

		D1(
			printf ("%s: %-20s: cmd_conf_data\n", prompt, __FUNCTION__)
		);
		
		dat->data[(sd->cmdlen - sizeof(struct jvisu_cmdhdr))] = '\0';
#if 0		
		if (jops->save_configs)
			jops->save_configs (dat->hdr.type, dat->data);
#endif
		break;
	}
	
	/* write a variable value */
	case cmd_send_var: {
				   
		struct jvisu_sndvar *dat = (struct jvisu_sndvar *) sd->buf;
		
		D1(
			printf("%s: %-20s: cmd_send_var\n", prompt, __FUNCTION__)
		);

		write_variable(pv_engine, dat, sd);
		break;
	}
			   
	default:
		break;
	}

	sd->status = 0;
	sd->bufpoi = 0;	
	sd->cmdlen = 0;

	return res;	
}

/** 
 *	Process requests
 *	
 *	@param sd	jvisu socket description structure
 *	@return		0 on success, else error number
 */
static int process_request (pv_engine_t *pv_engine, struct jvisu_sock *sd)
{
	int	res; 
	size_t	len = sizeof (int);

	/* FIXME: is this correct ? */
	if (sd->status == STAT_WRITE)
		return 0;

	/* check if socket is alive */
	res = read (sd->sd, &sd->buf[sd->bufpoi], 1);
	if (res <= 0) {
		shutdown_sock(pv_engine, sd);
		return 1;
	}
	sd->bufpoi++;
	
	len = sizeof (struct jvisu_cmdhdr);
	
	/* loop until buffer is empty or returns EOF */
	while (1) {	
		if (sd->status == STAT_READ) {
			if (!sd->cmdlen)
				len -= sd->bufpoi;
			else
				len = sd->cmdlen - sd->bufpoi;	
		}
		/* set status to READ */
		sd->status = STAT_READ;	

		/* check for more */
		res = read (sd->sd, &sd->buf[sd->bufpoi], len);
		
		/* handle EOF */
		if (res == EOF)
			return 0;

		/* socket error? */
		if (res < 0) {
			shutdown_sock(pv_engine, sd);
			return 1;
		}

		/* adjust buffer pointer */
		sd->bufpoi += res;
		
		/* have we read the commandlength already ? */
		if (!sd->cmdlen) {
			if (sd->bufpoi < sizeof (struct jvisu_cmdhdr))
				continue;
			sd->cmdlen = (unsigned int) sd->buf[2];
			sd->cmdlen |= ((unsigned int) sd->buf[3]) << 8;
			D1 (printf ("%s: %-20s: command: len=%d bytes\n", prompt, __FUNCTION__, sd->cmdlen));
		}
		
		/* have we got all bytes ? */
		if (sd->bufpoi == sd->cmdlen) {
			D1 (
				printf ("%s: %-20s: command: analysing (%d bytes read)\n", 
				prompt, __FUNCTION__, sd->cmdlen + sizeof (struct jvisu_cmdhdr))
			);
			return analyse_command(pv_engine, sd);
		}	
	}
	return 0; /* keep gcc happy */
}

/** 
 * 	Update fdset
 */
static void check_fdset (pv_engine_t *pv_engine)
{
	int	i;

	/* check server socket first */
	if (FD_ISSET(pv_engine->jv_active_sock[0].sd, &pv_engine->jv_read_fdset))
		do_accept(pv_engine, pv_engine->jv_active_sock[0].sd);
		
	/* server is always set */
	FD_SET(pv_engine->jv_active_sock[0].sd, &pv_engine->jv_read_fdset);

	/* loop through active connections
	 * 0 is always the server socket
	 */
	for (i = 1; i <= MAX_CONNECTIONS; i++) {

		if (pv_engine->jv_active_sock[i].sd == -1) 
			continue;
		
		if (pv_engine->jv_active_sock[i].status == STAT_WRITE) {
			if (FD_ISSET(pv_engine->jv_active_sock[i].sd, &pv_engine->jv_write_fdset)) {
				pv_engine->jv_active_sock[i].status &= ~STAT_WRITE;
			}
			FD_SET (pv_engine->jv_active_sock[i].sd, &pv_engine->jv_write_fdset);
		}
			
		if (FD_ISSET(pv_engine->jv_active_sock[i].sd, &pv_engine->jv_read_fdset)) {
			if (process_request (pv_engine, &pv_engine->jv_active_sock[i]))
				continue;
		}
		FD_SET (pv_engine->jv_active_sock[i].sd, &pv_engine->jv_read_fdset);
	}
}


/**
 *	start/stop JVisu
 *
 *	@param stop	set to 1 if you want to stop JVisu
 *	@return		0 on success
 */
int pv_stop_jvserver (pv_engine_t *pv_engine, int stop)
{
	pv_engine->jvisu_stop = stop;
	return 0;
}




/* 
 *  xml callbacks 
 */

static void XMLCALL start_element(void *userData, const char *name, const char **atts)
{
	pv_engine_t *pv_engine = (pv_engine_t*)userData;
	pv_t pv;
	
	if (strcmp(name,"pv") == 0) {

		if (pv_engine->pv_list == NULL)
			pv_engine->pv_list = list_init();

		pv_engine->pv_list = 
			list_add_tail(pv_engine->pv_list, (void*)&pv, sizeof(pv_t));
	}

	if ((strcmp(name,"name")    == 0) ||
	    (strcmp(name,"type")    == 0) ||
	    (strcmp(name,"offset")  == 0) ||
	    (strcmp(name,"desc")    == 0) ||
	    (strcmp(name,"default") == 0) ||
	    (strcmp(name,"scale")   == 0) ||
	    (strcmp(name,"unit")    == 0))
		pv_engine->xml_in_tag = 1;
	
	pv_engine->xml_depth += 1;
}

static void XMLCALL end_element(void *userData, const char *name)
{
	pv_engine_t *pv_engine = (pv_engine_t*)userData;
	pv_t *pv = (pv_t*)list_get_element_data(pv_engine->pv_list);

	if (strcmp(name,"pv") == 0)
		goto exit_end;

	if (strcmp(name,"name") == 0) {
		pv->name = malloc(pv_engine->xml_str_buf_len+1);
		if (!pv->name) {
			fprintf(stderr, "could not allocate memory\n");
			goto exit_end;
		}
		memcpy(pv->name,pv_engine->xml_str_buf,pv_engine->xml_str_buf_len);
		*(pv->name+pv_engine->xml_str_buf_len) = '\0';
		pv_engine->xml_in_tag = 0;
	}
	if (strcmp(name,"type") == 0) {
		pv->type = pv_type_undef;
		*(pv_engine->xml_str_buf+pv_engine->xml_str_buf_len) = '\0';
		if (strncmp(pv_engine->xml_str_buf,"uint8_t",XML_STR_BUF_SIZE) == 0)
			pv->type = pv_type_uint8;
		if (strncmp(pv_engine->xml_str_buf,"uint16_t",XML_STR_BUF_SIZE) == 0)
			pv->type = pv_type_uint16;
		if (strncmp(pv_engine->xml_str_buf,"uint32_t",XML_STR_BUF_SIZE) == 0)
			pv->type = pv_type_uint32;
		pv_engine->xml_in_tag = 0;
	}
	if (strcmp(name,"offset") == 0) {
		*(pv_engine->xml_str_buf+pv_engine->xml_str_buf_len) = '\0';
		pv->offset = strtoul(pv_engine->xml_str_buf, NULL, 0);
		if (errno) {
			fprintf(stderr, "error: %s must be unsigned long\n", pv_engine->xml_str_buf);
			return;
		}
		pv_engine->xml_in_tag = 0;
	}
	if (strcmp(name,"desc") == 0) {
		pv->desc = malloc(pv_engine->xml_str_buf_len+1);
		if (!pv->desc) {
			fprintf(stderr, "could not allocate memory\n");
			goto exit_end;
		}
		memcpy(pv->desc,pv_engine->xml_str_buf,pv_engine->xml_str_buf_len);
		*(pv->desc+pv_engine->xml_str_buf_len) = '\0';
		pv_engine->xml_in_tag = 0;
	}
	if (strcmp(name,"default") == 0) {
		/* FIXME */
		/* void            *dflt; */
		pv_engine->xml_in_tag = 0;
	}
	if (strcmp(name,"scale") == 0) {
		/* FIXME enum pv_scale   scale; */
		pv_engine->xml_in_tag = 0;
	}
	if (strcmp(name,"unit") == 0) {
		/* FIXME enum pv_unit unit; */
		pv_engine->xml_in_tag = 0;
	}

exit_end:
	pv_engine->xml_str_buf_len = 0;
	pv_engine->xml_depth -= 1;
}

static void XMLCALL character_data_handler(void *userData, const XML_Char *s, int len)
{
	pv_engine_t *pv_engine = (pv_engine_t*)userData;
	int rest, count;

	if (pv_engine->xml_in_tag) {
		rest = XML_STR_BUF_SIZE-1 - pv_engine->xml_str_buf_len;
		count = MIN(len,rest);
		memcpy(pv_engine->xml_str_buf+pv_engine->xml_str_buf_len,s,count);
		pv_engine->xml_str_buf_len += count;

		if (count < len) {
			fprintf(stderr,"warning: string length too long, truncating\n");
		}
	}
}

/*
 *  internal functions
 */

static void show_help(char *name)
{
	fprintf(stderr, "\nusage:   "
		"%s [-v] -d <file> -s <file> [-j <port>]\n\n"
		"  -v, --verbose             Verbose mode of operation\n"
		"  -d, --description <file>  Description XML File: specifies layout of the \n"
		"                            process variable store file\n"
		"  -s, --store <file>        Store file: this file holds the process variables\n"
		"  -j, --jvisu <port>        Start JVisu Socket Server\n"
		"\n", name);
}

/*
 *  Give JVisu pointers to a process variable
 */
static int get_symbol (pv_engine_t *pv_engine, struct jvisu_var *var, char *symbol)
{
	pv_t *pv;
	int finished = 0; 

	D1(printf("%s: %-20s: JVisu access to symbol <%s>\n", prompt, __FUNCTION__, symbol));

	if (!var || !symbol) {
		printf ("%s: %-20s: error: JVisu access failed\n", prompt, __FUNCTION__);
		return -EINVAL;
	}

	/*
	 * Roadmap: 
	 *
	 * * return -EINVAL if access is invalid
	 *
	 * * val->permissions = JVISU_PERM_RW | JVISU_PERM_RO | JVISU_PERM_WO
	 * * var->data        = (void *)(pointer_to_cell)
	 * * var->inttype     = jv_int, jv_short, ... && return 0
	 *                      return -EINVAL for illegal var type
	 * 
	 */
	
	var->data = NULL; 

	if (is_empty(pv_engine->pv_list))
		return -EINVAL;
	
	for (goto_head(&pv_engine->pv_list); !finished; goto_next(&pv_engine->pv_list)) {

		pv = (pv_t*)list_get_element_data(pv_engine->pv_list);
		
		if (is_last(pv_engine->pv_list))
			finished=1;

		if (!strncmp(symbol, pv->name, MIN(sizeof(symbol), sizeof(pv->name)))) {
			
			/* gotcha! */
			D1( printf("%s: %-20s: gotcha!\n", prompt, __FUNCTION__) );
			
			switch (pv->type) {
			case pv_type_uint8: 
				var->inttype = jv_byte;
				break; 
			case pv_type_uint16:
				var->inttype = jv_short;
				break;
			case pv_type_uint32:
				var->inttype = jv_int; 
			default:
				return -EINVAL;
			}
			
			var->data = (void*)(((uint8_t*)(pv_engine->pv_map))+pv->offset);

			var->permissions = JVISU_PERM_RW; 
			
			break; 
		}
	}

	/* not found? */
	if (!var->data)
		return -EINVAL;

	return 0;
}

/* FIXME: do I have to lock the pv_engine context? */

static void *pv_jvsocket_task (void *arg)
{
	pv_engine_t *pv_engine = (pv_engine_t*)arg;
	int res, fd, sd, current = 0;
	struct timeval timeo;

	/* initialize variable list */
//	INIT_LIST_HEAD (&pv_engine->jv_varlist);

	/* open bind and listen on socket */
	sd = start_socket (pv_engine, pv_engine->jv_port);
	if (sd < 0) {
		fprintf (stderr, "%s: %-20s: cannot open socket (port: %d)\n", prompt, __FUNCTION__, pv_engine->jv_port);
		goto exit_jvsocket_task;
	}

	D1(
		printf ("%s: %-20s: active (sd=%d), waiting for connections on port %d\n", 
		prompt, __FUNCTION__, sd, pv_engine->jv_port)
	);
	
	/* preset active SDs with -1 */
	memset (&pv_engine->jv_active_sock, 0xff, sizeof (pv_engine->jv_active_sock));

	/* store active socket */
	pv_engine->jv_active_sock[0].sd = sd;
	
	FD_ZERO(&pv_engine->jv_read_fdset);
	FD_ZERO(&pv_engine->jv_write_fdset);
	
	/* server is always set */
	FD_SET(sd, &pv_engine->jv_read_fdset);

	/* initialize timekeeping */
	gettimeofday(&pv_engine->jv_tact, NULL);
	
	while (!pv_engine->jvisu_stop) {

		/* check if there are active connections */
		fd = 0;
		for (current = 0; current <= MAX_CONNECTIONS; current++) {
			if (pv_engine->jv_active_sock[current].sd != -1)
				fd = MAX(pv_engine->jv_active_sock[current].sd, fd);
		}
		
		/* 
		 * if no connections -> come back once a second to check 
		 * pvengine->jvisu_stop; else use the given scan interval 
		 * for variable changes
		 */
		timeo.tv_sec = pv_engine->jv_connections ? 0 : 1;
		timeo.tv_usec = pv_engine->jv_connections ? (JVISU_INTERVAL * 1000) : 0;
		res = select(fd + 1, &pv_engine->jv_read_fdset, &pv_engine->jv_write_fdset, NULL, &timeo);

		/* FIXME: can we really continue ? */
		if (res < 0) {
			if (errno == EINTR) continue;
			fprintf (stderr, "%s: %s: error (errno=%i) on select: %d\n", prompt, __FUNCTION__, errno, res);
			continue;
		}

		/* check fdsets */
		check_fdset(pv_engine);				

		/* check variables */
		check_variables(pv_engine);
	}

exit_jvsocket_task: 
	
	D1(
		printf ("%s: %-20s: server stopped, shut down sockets\n", 
		prompt, __FUNCTION__)
	);	

	for (current = 0; current <= MAX_CONNECTIONS; current++) {
		if (pv_engine->jv_active_sock[current].sd != -1) 
			shutdown_sock (pv_engine, &pv_engine->jv_active_sock[current]);
	}		

	pthread_exit (NULL);
	return NULL;
}


/*
 *  API functions
 */

pv_engine_t *pv_init(int argc, char *argv[])
{
	int pv_store_size; 
	pv_t *tmp_pv = NULL; 
	
	/* xml parser */
	XML_Parser xml_parser;
	int xml_done = 0;
	char buf[1024];

	/* option parser */
	struct option long_options[] = {
		{ "help",        no_argument,       0, 'h' },
		{ "verbose",     no_argument,       0, 'v' },
		{ "description", required_argument, 0, 'd'}, 
		{ "store",       required_argument, 0, 's'},
		{ 0,             0,                 0,  0 },
	};
	int longindex;
	int opt;

	/* create and initialize new context */
	pv_engine_t *pv_engine = calloc(1,sizeof(pv_engine_t));
	if (pv_engine == NULL) {
		fprintf(stderr, "%s: %s: error: out of memory\n", prompt, __FUNCTION__);
		return NULL;
	}
	pv_engine->jv_varlist = NULL;

	/* parse command line switches */
	while ((opt = getopt_long(argc, argv, "hvd:s:j:", long_options, &longindex)) != -1) {
		switch (opt) {
		case 'h':
			show_help(basename(argv[0]));
			free(pv_engine);
			return NULL;
		case 'v':
			pv_verbose=1;
			break;	
		case 'd':
			pv_engine->xml_filename=optarg;
			break;
		case 's':
			pv_engine->pv_filename=optarg;
			break;
		case 'j':
			pv_engine->jv_port=strtoul(optarg,NULL,0);
			break;
                default:
                        show_help(argv[0]);
			free(pv_engine);
			return NULL;
                }
        }

	/* parse command line options */
	if ( ((argc-optind) != 0) ||
	     (pv_engine->xml_filename == NULL) || 
	     (pv_engine->pv_filename == NULL)
	) {
		show_help(argv[0]);
		free(pv_engine);
		return NULL;
	}

	if (pv_verbose) {
		printf("\n");
		printf("description file...: %s\n", pv_engine->xml_filename);
		printf("store file.........: %s\n", pv_engine->pv_filename);
		printf("\n");
	}
	
	/* analyse xml description file */
	pv_engine->xml_file = fopen(pv_engine->xml_filename, "r");
	if (!pv_engine->xml_file) {
		fprintf(stderr, "could not open xml file \"%s\": ", pv_engine->xml_filename);
		perror("");
		free(pv_engine);
		return NULL;
	}
	
	xml_parser = XML_ParserCreate(NULL);
	XML_SetUserData(xml_parser, pv_engine);
	XML_SetElementHandler(xml_parser, start_element, end_element);
	XML_SetCharacterDataHandler(xml_parser, character_data_handler);

	do {
		size_t len = fread(buf, 1, sizeof(buf), pv_engine->xml_file);
		xml_done = len < sizeof(buf);
		if (XML_Parse(xml_parser, buf, len, xml_done) == XML_STATUS_ERROR) {
			fprintf(stderr, "%s at line %d\n",
				XML_ErrorString(XML_GetErrorCode(xml_parser)),
				XML_GetCurrentLineNumber(xml_parser));
			fclose(pv_engine->xml_file);
			free(pv_engine);
			return NULL;
		}

	} while (!xml_done);

	XML_ParserFree(xml_parser);
	fclose(pv_engine->xml_file);

	/* tell us which process variables have been found */
	if ((is_empty(pv_engine->pv_list)) && pv_verbose) {
		
		pv_t *pv;
		int finished = 0; 

		printf("process variables:\n");
		printf("------------------\n\n");
		
		for (goto_head(&pv_engine->pv_list); !finished; goto_next(&pv_engine->pv_list)) {
			pv = (pv_t*)list_get_element_data(pv_engine->pv_list);
			
			if (is_last(pv_engine->pv_list))
				finished=1;

			printf("n:%s o=%li\n",pv->name, pv->offset);
		}
	}

	/* open & mmap process variable space */
	pv_engine->pv_file = open(pv_engine->pv_filename, O_RDWR);
	if (pv_engine->pv_file < 0) {
		fprintf(stderr, "error: could not open process variable backing store\n");
		exit(EXIT_FAILURE);
	}

	/* find out size of pv area */
	pv_store_size=0; 
	if (is_last(pv_engine->pv_list) != -1) {
	
		int finished = 0; 

		for (goto_head(&pv_engine->pv_list); !finished; goto_next(&pv_engine->pv_list)) {
			
			pv_t *pv = (pv_t*)list_get_element_data(pv_engine->pv_list);
			
			if (is_last(pv_engine->pv_list))
				finished=1;
			
			if (pv->offset > pv_store_size) {
				pv_store_size = pv->offset;
				tmp_pv = pv; 
			}
		}
	}

	switch (tmp_pv->type) {
		case pv_type_uint8:  pv_store_size += 1; break;
		case pv_type_uint16: pv_store_size += 2; break;
		case pv_type_uint32: pv_store_size += 4; break; 
		default: 
			fprintf(stderr, "error, unknown type\n");
			exit(EXIT_FAILURE);
	}

	if (pv_verbose) 
		printf("pv store size = %i\n", pv_store_size); 

	pv_engine->pv_map = mmap(NULL, pv_store_size, PROT_READ|PROT_WRITE, MAP_SHARED, pv_engine->pv_file, 0);
	if (pv_engine->pv_map < 0) {
		fprintf(stderr, "error: could not mmap\n");
		exit(EXIT_FAILURE);
	}

	/* locking */
	/* FIXME: use absolute paths */
	pv_engine->pv_sem_key = ftok(pv_engine->pv_filename, 42); 
	pv_engine->pv_sem = semget(pv_engine->pv_sem_key, 1, 0666 | IPC_CREAT);
	if (pv_engine->pv_sem < 0) {
		fprintf(stderr, "error: could not create semaphore\n");
		perror("error: ");
		exit(EXIT_FAILURE);
	}
	{
		union semun sem_val; 
		sem_val.val = 1; 
		if (semctl(pv_engine->pv_sem, 0, SETVAL, sem_val)) {
			fprintf(stderr, "error: could not initialize semaphore\n");
			perror("error: ");
			exit(EXIT_FAILURE);
		}
	}
	if (pv_verbose)
		fprintf(stdout, "\n%s: %s: sem_id=%i\n", prompt, __FUNCTION__, pv_engine->pv_sem);

	if (pv_engine->jv_port != 0) {
	
		/* create jvisu socket thread; this may not work if another process has already started one */
		if (pv_verbose)
			fprintf(stdout, "%s: %s: starting JVisu socket server on port %i\n", prompt, __FUNCTION__, pv_engine->jv_port);
		if (pthread_create(&pv_engine->jvsocket_thread, 0, pv_jvsocket_task, (void*)pv_engine) != 0) {
			fprintf(stderr, "warning: could not create socket thread\n");
		}

	} else 
		if (pv_verbose)
			fprintf(stdout, "%s: %s: -j not given, so don't start JVisu socket server\n", prompt, __FUNCTION__);	
	
	if (pv_verbose)
		printf("\n");
	
	return pv_engine;
}

void pv_finish (pv_engine_t *pv_engine)
{
	/* shutdown jvisu socket server */
	D1(printf("%s: %-20s: stopping server\n", prompt, __FUNCTION__));
	pv_stop_jvserver(pv_engine, 1);
	
	if (pv_engine->jvsocket_thread) {
		D1(printf("%s: %-20s: waiting for server thread\n", prompt, __FUNCTION__));
		pthread_join(pv_engine->jvsocket_thread, 0);
		D1(printf("%s: %-20s: joined\n", prompt, __FUNCTION__));
	}

	/* FIXME: do we need to remove the SysV semaphore? */

	/* shut down backing store */
	close(pv_engine->pv_file);

	/* remove process variables */
	if (pv_verbose)
		printf("freeing pv list:");

	if (is_last(pv_engine->pv_list) != -1) {

		int finished = 0; 
	
		for (goto_head(&pv_engine->pv_list); !finished; goto_next(&pv_engine->pv_list)) {
			
			pv_t *pv = (pv_t*)list_get_element_data(pv_engine->pv_list);
			
			if (is_last(pv_engine->pv_list))
				finished=1;

			if (pv == NULL) {
				fprintf(stderr, "error: trying to free nonexistent list element\n");
				exit(EXIT_FAILURE);
			}
			free(pv->name);
			free(pv->desc);
			if (pv_verbose)
				printf(".");
			
			if (is_last(pv_engine->pv_list))
				goto end_finish;

			pv_engine->pv_list = list_del_element(pv_engine->pv_list);
		}
	}
	
end_finish:	
	if (pv_verbose)
		printf("\n");

	return;
}


uint8_t *pv_register_uint8(pv_engine_t *pv_engine, const char *varname)
{
	pv_t *pv;
	uint8_t *ret = NULL;
	int finished = 0; 

	if (is_last(pv_engine->pv_list) == -1) 
		return NULL; 
	
	for (goto_head(&pv_engine->pv_list); !finished; goto_next(&pv_engine->pv_list)) {

		pv = (pv_t*)list_get_element_data(pv_engine->pv_list);
			
		if (is_last(pv_engine->pv_list))
			finished=1;

		if (!strncmp(varname, pv->name, MIN(sizeof(varname), sizeof(pv->name)))) {
			ret = (uint8_t*)(((uint8_t*)(pv_engine->pv_map))+pv->offset);
			break; 
		}
	}

	return ret;
}


uint16_t *pv_register_uint16(pv_engine_t *pv_engine, const char *varname)
{
	pv_t *pv;
	uint16_t *ret = NULL;
	int finished = 0; 

	if (is_last(pv_engine->pv_list) == -1) 
		return NULL; 
	
	for (goto_head(&pv_engine->pv_list); !finished; goto_next(&pv_engine->pv_list)) {
		
		pv = (pv_t*)list_get_element_data(pv_engine->pv_list);
		
		if (is_last(pv_engine->pv_list))
			finished=1;
		
		if (!strncmp(varname, pv->name, MIN(sizeof(varname), sizeof(pv->name)))) {
			ret = (uint16_t*)(((uint8_t*)(pv_engine->pv_map))+pv->offset);
			break; 
		}
	}

	return ret;
}


uint32_t *pv_register_uint32(pv_engine_t *pv_engine, const char *varname)
{
	pv_t *pv;
	uint32_t *ret = NULL;
	int finished = 0; 

	if (is_last(pv_engine->pv_list) == -1) 
		return NULL; 
	
	for (goto_head(&pv_engine->pv_list); !finished; goto_next(&pv_engine->pv_list)) {
		
		if (is_last(pv_engine->pv_list))
			finished=1;

		pv = (pv_t*)list_get_element_data(pv_engine->pv_list);
		if (!strncmp(varname, pv->name, MIN(sizeof(varname), sizeof(pv->name)))) {
			ret = (uint32_t*)(((uint8_t*)(pv_engine->pv_map))+pv->offset);
			break; 
		}
	}

	return ret;
}

